use rand::rngs::StdRng;
use rand::RngCore;
use rand::SeedableRng;
use rug::{Assign, Integer};

pub fn encrypt_bit(bit: u8, key: &Integer) -> Integer {
    assert!(bit == 0 || bit == 1, "Only bits (0 or 1) are allowed");

    let mut rng = StdRng::from_os_rng();

    let q: Integer = Integer::from(rng.next_u64()) % key;

    // noise
    let p_div_4: Integer = key.clone() / 4;
    let r = std::iter::repeat_with(|| Integer::from(rng.next_u64()))
        .find(|r| r.clone() * 2 < p_div_4)
        .unwrap();

    // c = b + p*q + 2*r
    let mut c = Integer::new();
    c.assign(key.clone() * &q + 2 * &r + bit);

    c
}

pub fn decrypt_bit(encrypted: &Integer, key: &Integer) -> u8 {
    let bit: Integer = (encrypted.clone() % key.clone()) % 2; // ((c mod p) mod 2)
    bit.to_u8().unwrap()
}

pub fn generate_secret_key(eta: u32) -> Integer {
    let mut rng = StdRng::from_os_rng();
    let lower = Integer::from(1) << (eta - 1);
    let upper = Integer::from(1) << eta;
    let range: Integer = upper - &lower;
    std::iter::repeat_with(|| Integer::from(rng.next_u64()) % &range + &lower)
        .find(|p: &Integer| p.is_odd())
        .unwrap()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_secret_key_generation_properties() {
        let eta = 64;
        let key = generate_secret_key(eta);
        assert!(key.significant_bits() <= eta);
        assert!(key.significant_bits() >= eta - 1);
        assert!(key.is_odd());
    }

    #[test]
    fn test_encrypt_decrypt_bit() {
        let eta = 64;
        let key = generate_secret_key(eta);

        let bit = 0;
        let ciphertext = encrypt_bit(bit, &key);
        let decrypted = decrypt_bit(&ciphertext, &key);
        assert_eq!(decrypted, bit);

        let bit = 1;
        let ciphertext = encrypt_bit(bit, &key);
        let decrypted = decrypt_bit(&ciphertext, &key);
        assert_eq!(decrypted, bit);
    }

    #[test]
    fn test_multiple_encryptions_are_different() {
        let eta = 64;
        let key = generate_secret_key(eta);

        let bit = 1;
        let c1 = encrypt_bit(bit, &key);
        let c2 = encrypt_bit(bit, &key);

        assert_ne!(c1, c2);
        assert_eq!(decrypt_bit(&c1, &key), bit);
        assert_eq!(decrypt_bit(&c2, &key), bit);
    }

    #[test]
    fn test_is_homomorphic() {
        let eta = 1024;
        let key = generate_secret_key(eta);

        for &(b1, b2) in &[(0, 0), (0, 1), (1, 0), (1, 1)] {
            let c1 = encrypt_bit(b1, &key);
            let c2 = encrypt_bit(b2, &key);

            let sum: Integer = c1.clone() + &c2;
            let product: Integer = c1 * c2;

            let decrypted_sum = decrypt_bit(&sum, &key);
            let decrypted_product = decrypt_bit(&product, &key);

            let expected_xor = b1 ^ b2;
            let expected_and = b1 & b2;

            assert_eq!(
                decrypted_sum, expected_xor,
                "Failed XOR test for inputs: {}, {}",
                b1, b2
            );
            assert_eq!(
                decrypted_product, expected_and,
                "Failed AND test for inputs: {}, {}",
                b1, b2
            );
        }
    }
}
